# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class ProblemDecomposition(abc.ABC):

    def __init__(self):
        self.electronic_structure_solver = None
        self.electron_localization_method = None

    @abc.abstractmethod
    def simulate(self, molecule, fragment_atoms, mean_field=None):
        pass
